/*
Copyright (c) 2013-2015 Gustav Hartvigsson

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

#include "Callback.h"

SCallbackEntry *
s_callback_entry_new_real (const schar * name,
                           Callback callback,
                           SCallbackType type) {
  SCallbackEntry * self = s_malloc (sizeof (SCallbackEntry));
  self->name = s_string_new (name);
  self->callback = callback;
  self->type = type;
  return self;
}

void
s_callback_entry_free (SCallbackEntry * self) {
  s_free (self->name);
  s_free (self);
}

void
s_object_install_callbacks (SObject * obj,
                            size_t n_callbacks,
                            SCallbackEntry ** callback_entries) {
  assert (obj);
  assert (callback_entries);
  s_dbg_print ("Adding Callbacks, n_callbacks: %zd", n_callbacks);
  for (size_t i = 0; i < n_callbacks ; i++) {
    s_dbg_print ("Loop: Adding Callbacks");
    SCallbackEntry * entry = callback_entries[i];
    s_object_install_callback (obj, entry);
  }
}

void
s_object_install_callback (SObject * obj,
                           SCallbackEntry * callback_entry) {
  //
  assert (obj);
  assert (callback_entry);
  s_dbg_print ("Adding callback");
  s_map_add (obj->callbacks, s_string_new (callback_entry->name), callback_entry);
}

spointer
s_object_notify (SObject * self,
                 schar * name,
                 spointer * user_data) {
  //
  SCallbackEntry * entry = s_map_get (self->callbacks, name);
  assert (s_string_is_equal (entry->name, name));

  Callback callback = entry->callback;
  return callback (self, user_data);
}

