#include "mm.h"
#include "mm_mark_and_sweep.h"
#include "utils.h"

/*
Copyright (c) 2013-2016 Gustav Hartvigsson

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
 */

static sboolean _s_mm_initialized = FALSE;
static SMMType _s_mm_type = S_MM_TYPE_NONE;

void
s_mm_init (SMMType type) {
  if (!_s_mm_initialized) {
    _s_mm_initialized = TRUE;
    _s_mm_type = type;
    if (type == S_MM_TYPE_MARK_AND_SWEEP) {
      s_mm_mark_and_sweep_init ();
    }
  }
}

schar *
s_mm_type_get_name (SMMType type) {
  return SMMTypeName[type];
}

void
s_mm_cleanup () {
  if (_s_mm_type == S_MM_TYPE_MARK_AND_SWEEP) {
    s_mm_mark_and_sweep_cleanup ();
  }
}

SMMType
s_mm_get_type () {
  return _s_mm_type;
}

spointer
s_malloc (size_t size) {
  #if USE_GC
  if (_s_mm_type == S_MM_TYPE_GC) {
    return GC_MALLOC (size);
  }
  #endif
  return malloc (size);
}

void
s_free (spointer pointer) {
  switch (_s_mm_type) {
    case (S_MM_TYPE_GC):
    #if USE_GC
      pointer = NULL;
      break;
    #endif
    case (S_MM_TYPE_NONE):
    case (S_MM_TYPE_LIBC):
      free (pointer);
      break;
    case (S_MM_TYPE_MARK_AND_SWEEP):
      s_mm_mark_and_sweep_add_pointer_to_list (pointer);
      break;
    default:
      s_err_print ("Reaching a place that should not be reached.");
  }
}

spointer
s_realloc (spointer pointer, size_t size) {
  #if USE_GC
  if (_s_mm_type == S_MM_TYPE_GC) {
    return GC_REALLOC (pointer, size);
  }
  #endif
  return realloc (pointer, size);
}

spointer
s_calloc (size_t num, size_t size) {
  #if USE_GC
  if (_s_mm_type == S_MM_TYPE_GC) {
    /* This should work. */
    spointer ret_val = GC_MALLOC (num * size);
    return ret_val;
  }
  #endif
  return calloc (num, size);
}
