#include "../test_macros.h"
#include "../../libssts/SimpleTypeSystem.h"

typedef struct MyState MyState;
typedef struct MyTestObject MyTestObject;

sboolean _my_allocator (MyState *);
sboolean _my_deallocator (MyState *);

void _my_state_free (MyState *);
MyState * _my_state_new (int);

MyTestObject * _my_test_object_new ();
void _my_test_object_free (MyTestObject *);

struct MyState {
  int itt;
  int max_itt;
  SLinkedList * arr;
};

struct MyTestObject {
  int foo;
  char bar;
  int foobar;
  long barfoo;
};

MyState *
_my_state_new (int max_itt) {
  MyState * self = s_malloc (sizeof (MyState));
  self->itt = 0;
  self->max_itt = max_itt;
  self->arr = s_linked_list_new (FREEFUNC(_my_test_object_free));
  return self;
}

void
_my_state_free (MyState * self) {
  s_print ("Running _s_state_free\n");
  s_linked_list_free (self->arr, TRUE);
  s_free (self);
}

MyTestObject *
_my_test_object_new () {
  MyTestObject * self = s_malloc (sizeof (MyTestObject));
  return self;
}

void
_my_test_object_free (MyTestObject * self) {
  s_free (self);
}



sboolean
_my_allocator (MyState * state) {
  s_print ("Running alloc Worker... Itt: %d\n", state->itt);
  if (state->itt >= state->max_itt) {
    // Do not continue worker
    return FALSE;
  }
  MyTestObject * obj = s_malloc (sizeof (MyTestObject));
  
  s_linked_list_append (state->arr, obj);
  state->itt = state->itt + 1;
  // continue worker
  return TRUE;
}

sboolean
_my_dealloctater (MyState * state) {
  s_print ("Running dealloc Worker... Itt: %d\n", state->itt);
  // The allorator has not been run.
  if (state->itt == 0) {
    // continue worker
    return TRUE;
  }
  if (s_linked_list_len (state->arr) != 0) {
    s_linked_list_remove_current (state->arr, TRUE);
    // continue worker
    return TRUE;
  }
  // Stop worker
  return FALSE;
}

spointer
_my_state_teardown (MyState * self, spointer * user_data) {
  _my_state_free (self);
  return NULL;
}

int
main (int argc, char ** argv) {
  s_mm_init (S_MM_TYPE_GC);
  
  // Get the main loop before we run it.
  SMainLoop * loop = s_main_loop_get_default ();
  
  // create a state object to track stuffs.
  MyState * state = _my_state_new (30);
  
  // Add teardown handler
  s_main_loop_add_teardown_handler (loop,
                                    CALLBACK(_my_state_teardown),
                                    state,
                                    NULL);
  
  /* Add workers.
   * Allocator on ring one,
   * De-allorator in ring two
   */
  s_main_loop_add_worker (loop,
                          S_MAIN_LOOP_RING_ONE,
                          RUNFUNC(_my_allocator),
                          state);
  s_main_loop_add_worker (loop,
                          S_MAIN_LOOP_RING_TWO,
                          RUNFUNC(_my_dealloctater),
                          state);
  
  s_main_loop_run (loop);
  
  s_teardown ();
}
